#include <ultra/core.hpp>
#include <boost/math/quaternion.hpp>


struct js {
    Poco::Dynamic::Var var;
    template<typename T> js(vector<T> const& arg) { var = arg; }
    template<typename T, typename U> js(map<T, U> const& arg) {
        Poco::JSON::Object obj;
        for (auto const& p : arg) {
            obj.set(to_string(p.first), p.second);
        }
        var = obj;
    }
    template<typename U> js(map<string, U> const& arg) {
        Poco::JSON::Object obj;
        for (auto const& p : arg) {
            obj.set(p.first, p.second);
        }
        var = obj;
    }
};
ostream& operator << (ostream& out, js const& value)
{
    Poco::JSON::Stringifier::stringify(value.var, out);
    return out;
}


struct psk_header {
    char chunk_id[20];
    uint32_t type_flag;
    uint32_t data_size;
    uint32_t data_count;
};
ostream& operator << (ostream& out, psk_header const& header)
{
    return out << "{" << string(header.chunk_id) << "} items: " << header.data_count << ", size: " << header.data_size;
}

struct psk_point {
    float x, y, z;
};
ostream& operator << (ostream& out, psk_point const& pt)
{
    return out << "[" << pt.x << ", " << pt.y << ", " << pt.z << "]";
}

struct psk_wedge {
    uint16_t point_index;
    uint16_t padding;
    float u, v;
    uint8_t material_index;
    uint8_t reserved;
    uint16_t padding2;
};
ostream& operator << (ostream& out, psk_wedge const& wedge)
{
    return out << "point #" << wedge.point_index << ", [u,v] = [" << wedge.u << ", " << wedge.v << "], material #" << (int) wedge.material_index;
}

struct psk_face {
    uint16_t wedge0, wedge1, wedge2;
    uint8_t material_index, aux_material_index;
    uint32_t smoothing_groups;
};
ostream& operator << (ostream& out, psk_face const& face)
{
    return out << "wedges [" << face.wedge0 << ", " << face.wedge1 << ", " << face.wedge2 << "]" <<
        ", materials [" << (int) face.material_index << ", " << (int) face.aux_material_index << "]" <<
        ", smoothing groups: " << face.smoothing_groups;
}

struct psk_material {
    char name[64];
    uint32_t texture_index;
    uint32_t poly_flags;
    uint32_t aux_material;
    uint32_t aux_flags;
    uint32_t lod_bias;
    uint32_t lod_style;
};
ostream& operator << (ostream& out, psk_material const& material)
{
    return out << "material " << string(material.name) << ", texture index: "  << material.texture_index << ", poly flags: " << material.poly_flags <<
        ", aux material: " << material.aux_material << ", aux flags: " << material.aux_flags <<
        ", lod bias: " << material.lod_bias << ", lod style: " << material.lod_style;
}

struct psk_bone {
    char name[64];
    uint32_t flags;
    uint32_t num_children;
    uint32_t parent_index;
    float orientation[4];
    float position[3];
    float length;
    float size[3];
};
ostream& operator << (ostream& out, psk_bone const& bone)
{
    return out << "bone " << string(bone.name) << ", flags: " << bone.flags << ", children: " << bone.num_children << ", parent: " << bone.parent_index <<
        ", orientation: [" << bone.orientation[0] << ", " << bone.orientation[1] << ", " << bone.orientation[2] << ", " << bone.orientation[3] << "]" <<
        ", position: [" << bone.position[0] << ", " << bone.position[1] << ", " << bone.position[2] << "]" <<
        ", len: " << bone.length <<
        ", size: [" << bone.size[0] << ", " << bone.size[1] << ", " << bone.size[2] << "]";
}

struct psk_bone_weight {
    float weight;
    uint32_t point_index;
    uint32_t bone_index;
};
ostream& operator << (ostream& out, psk_bone_weight const& weight)
{
    return out << "for point #" << weight.point_index << " bone #" << weight.bone_index << " has weight=" << weight.weight;
}


typedef boost::math::quaternion<float> fquaternion;


template<typename TBoneParentFn, typename TBoneData>
void calculate_skelet(TBoneParentFn bone_parent_fn, TBoneData const& bones_source, vector<fquaternion>& abs_positions, vector<fquaternion>& abs_orientations)
{
    vector<float> result;
    int n = bones_source.size();
    abs_positions.resize(n);
    abs_orientations.resize(n);
    for (int i = 0; i < n; ++i) {
        fquaternion or{bones_source[i].orientation[3], bones_source[i].orientation[0], bones_source[i].orientation[1], bones_source[i].orientation[2]};
        fquaternion pos{0.0f, bones_source[i].position[0], bones_source[i].position[1], bones_source[i].position[2]};
        int iparent = bone_parent_fn(i);
        if (i == 0) {
            abs_orientations[i] = or;
            abs_positions[i] = pos;
        } else {
            auto const& por = abs_orientations[iparent];
            auto const& ppos = abs_positions[iparent];
            abs_orientations[i] = or * por;
            abs_positions[i] = ppos + boost::math::conj(por) * pos * por;
        }
        if (abs(abs_positions[i].R_component_1()) > 1e-5 || abs(abs(abs_orientations[i]) - 1.0) > 1e-5) {
            cerr << "WUT pos: " << abs(abs_positions[i].R_component_1()) << " or: " << abs(abs(abs_orientations[i]) - 1.0) << endl;
        }
    }
}


struct psa_animinfo {
    char			Name[64];				// Animation's name
    char			Group[64];				// Animation's group name
    int				TotalBones;				// TotalBones * NumRawFrames is number of animation keys to digest.
    int				RootInclude;			// 0 none 1 included (unused)
    int				KeyCompressionStyle;	// Reserved: variants in tradeoffs for compression.
    int				KeyQuotum;				// Max key quotum for compression; ActorX sets this to numFrames*numBones
    float			KeyReduction;			// desired
    float			TrackTime;				// explicit - can be overridden by the animation rate
    float			AnimRate;				// frames per second.
    int				StartBone;				// Reserved: for partial animations
    int				FirstRawFrame;			// global number of first animation frame
    int				NumRawFrames;			// NumRawFrames and AnimRate dictate tracktime...
};

struct psa_animkey {
    float position[3];
    float orientation[4];
    float time;
};

struct animation_data {
    int frames, bones;
    boost::iterator_range<psa_animkey *> animation;
};

struct psk_psa_reader {
    string psk_file, psa_file;

    boost::iterator_range<psk_point *> points;
    boost::iterator_range<psk_wedge *> wedges;
    boost::iterator_range<psk_face *> faces;
    boost::iterator_range<psk_bone *> bones;
    boost::iterator_range<psk_bone_weight *> bone_weights;
        
    boost::iterator_range<psa_animinfo *> animinfos;
    map<string, animation_data> animations;

    vector<string> psk_bone_names, psa_bone_names;
    vector<int> psa_to_psk_bone, psk_to_psa_bone;

    explicit psk_psa_reader(string const& psk_filename, string const& psa_filename)
    {
        psk_file = php::file_get_contents(psk_filename);
        char *ptr = &psk_file[0];
        while (ptr < &psk_file[0] + psk_file.length()) {
            ptr = read_psk_header(ptr);
        }
        psa_file = php::file_get_contents(psa_filename);
        ptr = &psa_file[0];
        while (ptr < &psa_file[0] + psa_file.length()) {
            ptr = read_psa_header(ptr);
        }
    }

    char *read_psk_header(char *buff)
    {
        psk_header *header = (psk_header *) buff;
        clog << *header << endl;
        string id = string(&header->chunk_id[0]);
        buff += sizeof(psk_header);
        if (id == "ACTRHEAD") {
        } else if (id == "PNTS0000") {
            points = boost::iterator_range<psk_point *>((psk_point *) buff, (psk_point *) buff + header->data_count);
            for (int i = 0; i < header->data_count; ++i, buff += header->data_size) {
                psk_point& pt = *(psk_point *) buff;
                //cout << pt << endl;
            }
        } else if (id == "VTXW0000") {
            wedges = boost::iterator_range<psk_wedge *>((psk_wedge *) buff, (psk_wedge *) buff + header->data_count);
            for (int i = 0; i < header->data_count; ++i, buff += header->data_size) {
                psk_wedge& wedge = *(psk_wedge *) buff;
                //cout << wedge << endl;
            }
        } else if (id == "FACE0000") {
            faces = boost::iterator_range<psk_face *>((psk_face *) buff, (psk_face *) buff + header->data_count);
            for (int i = 0; i < header->data_count; ++i, buff += header->data_size) {
                psk_face& face = *(psk_face *) buff;
                //cout << face << endl;
            }
        } else if (id == "MATT0000") {
            for (int i = 0; i < header->data_count; ++i, buff += header->data_size) {
                psk_material& material = *(psk_material *) buff;
                //cout << material << endl;
            }
        } else if (id == "REFSKELT") {
            bones = boost::iterator_range<psk_bone *>((psk_bone *) buff, (psk_bone *) buff + header->data_count);
            for (int i = 0; i < header->data_count; ++i, buff += header->data_size) {
                psk_bone& bone = *(psk_bone *) buff;
                psk_bone_names.push_back(bone.name);
                //clog << bone << endl;
            }
        } else if (id == "RAWWEIGHTS") {
            map<int, int> bone_counts;
            bone_weights = boost::iterator_range<psk_bone_weight *>((psk_bone_weight *) buff, (psk_bone_weight *) buff + header->data_count);
            for (int i = 0; i < header->data_count; ++i, buff += header->data_size) {
                psk_bone_weight& weight = *(psk_bone_weight *) buff;
                ++bone_counts[weight.point_index];
                //cout << weight << endl;
            }
            map<int, int> bones_stats;
            for (auto const& point_bones : bone_counts) {
                ++bones_stats[point_bones.second];
            }
            clog << js(bones_stats) << endl;
        } else {
            buff += header->data_count * header->data_size;
        }
        return buff;
    }

    char *read_psa_header(char *ptr)
    {
        psk_header& header = *(psk_header *) ptr;
        string id = string(header.chunk_id);
        ptr += sizeof(psk_header);
        clog << header << endl;
        if (id == "ANIMHEAD") {

        } else if (id == "BONENAMES") {
            char *p = ptr;
            for (int i = 0; i < header.data_count; ++i, p += header.data_size) {
                psa_bone_names.push_back(p);
            }

            // Bones order don't match in PSK and PSA
            map<string, int> psk_bone_to_id;
            for (int i = 0, ilen = psk_bone_names.size(); i < ilen; ++i) {
                psk_bone_to_id[psk_bone_names[i]] = i;
            }
            psa_to_psk_bone.resize(psa_bone_names.size());
            psk_to_psa_bone.resize(psk_bone_names.size());
            for (int i = 0, ilen = psa_bone_names.size(); i < ilen; ++i) {
                psa_to_psk_bone[i] = psk_bone_to_id[psa_bone_names[i]];
                psk_to_psa_bone[psa_to_psk_bone[i]] = i;
            }
        } else if (id == "ANIMINFO") {
            animinfos = boost::iterator_range<psa_animinfo *>((psa_animinfo *) ptr, (psa_animinfo *) ptr + header.data_count);
            for (int i = 0; i < header.data_count; ++i) {
                psa_animinfo& animinfo = ((psa_animinfo *) ptr)[i];
                clog << "animation " << string(animinfo.Name) << ", group " << string(animinfo.Group) <<
                    ": " << animinfo.TotalBones << " bones x " << animinfo.NumRawFrames << " frames at " << animinfo.AnimRate << " fps" << endl;
            }
        } else if (id == "ANIMKEYS") {
            for (int i = 0; i < header.data_count; ++i) {
                psa_animkey& animkey = ((psa_animkey *) ptr)[i];
                if (animkey.time != 1.0) {
                    cerr << animkey.time << endl;
                }
            }
            int frame = 0;
            psa_animkey *animkeys = (psa_animkey *) ptr;
            for (auto&& animinfo : animinfos) {
                auto& anim = animations[string(animinfo.Name)];
                anim.frames = animinfo.NumRawFrames;
                anim.bones = animinfo.TotalBones;
                anim.animation = boost::iterator_range<psa_animkey *>(animkeys, animkeys + animinfo.NumRawFrames * animinfo.TotalBones);
                animkeys += animinfo.NumRawFrames * animinfo.TotalBones;
            }
        } else {

        }
        ptr += header.data_count * header.data_size;
        return ptr;
    }


    vector<float> vertex_buffer;
    vector<float> texcoord_buffer;
    vector<int> index_buffer;

    map<string, vector<vector<float>>> bone_quat_buffer_frames_anims;
    map<string, vector<vector<float>>> bone_pos_buffer_frames_anims;

    vector<float> bone_weight_buffer;
    vector<vector<float>> bone_data_buffer;

    void prepare_buffers()
    {
        int N = wedges.size();

        // Per-vertex data
        for (int i = 0; i < N; ++i) {
            vertex_buffer.push_back(points[wedges[i].point_index].x);
            vertex_buffer.push_back(points[wedges[i].point_index].y);
            vertex_buffer.push_back(points[wedges[i].point_index].z);

            texcoord_buffer.push_back(wedges[i].u);
            texcoord_buffer.push_back(wedges[i].v);
        }

        // Index buffer
        for (auto const& face : faces) {
            index_buffer.push_back(face.wedge0);
            index_buffer.push_back(face.wedge1);
            index_buffer.push_back(face.wedge2);
        }

        // Bones info
        int P = points.size();
        vector<vector<int>> bones_idx(P);
        vector<vector<float>> w(P);
        for (auto const& weight : bone_weights) {
            w[weight.point_index].push_back(weight.weight);
            bones_idx[weight.point_index].push_back(weight.bone_index);
        }

        bone_weight_buffer.resize(4 * N);
        for (int i = 0, c = 0; i < N; ++i, c += 4) {
            copy(w[wedges[i].point_index].begin(), w[wedges[i].point_index].end(), bone_weight_buffer.begin() + c);
        }

        {
            // Calc default skelet
            vector<fquaternion> positions, orientations;
            calculate_skelet([&](int i) { return bones[i].parent_index; }, bones, positions, orientations);

            // Bones info
            bone_data_buffer.resize(4);
            for (int i = 0; i < 4; ++i) {
                bone_data_buffer[i].resize(4 * N);
                for (int j = 0; j < N; ++j) {
                    int bone = bones_idx[wedges[j].point_index].size() > i ? bones_idx[wedges[j].point_index][i] : 0;
                    auto& point_pos = points[wedges[j].point_index];
                    fquaternion position{0.0, point_pos.x, point_pos.y, point_pos.z};
                    auto rel_pos = orientations[bone] * (position - positions[bone]) * boost::math::conj(orientations[bone]);

                    bone_data_buffer[i][j * 4 + 0] = rel_pos.R_component_2();
                    bone_data_buffer[i][j * 4 + 1] = rel_pos.R_component_3();
                    bone_data_buffer[i][j * 4 + 2] = rel_pos.R_component_4();
                    bone_data_buffer[i][j * 4 + 3] = bone;
                }
            }

            // Default pose
            vector<float> quat, pos;
            for (int i = 0; i < bones.size(); ++i) {
                quat.push_back(orientations[i].R_component_2());
                quat.push_back(orientations[i].R_component_3());
                quat.push_back(orientations[i].R_component_4());
                quat.push_back(orientations[i].R_component_1());

                pos.push_back(positions[i].R_component_2());
                pos.push_back(positions[i].R_component_3());
                pos.push_back(positions[i].R_component_4());
                pos.push_back(0.0);
            }
            bone_quat_buffer_frames_anims["default"].push_back(quat);
            bone_pos_buffer_frames_anims["default"].push_back(pos);
        }

        // Animations
        for (auto const& name_animation : animations) {
            string const& name = name_animation.first;
            auto const& anim = name_animation.second;
            auto& quat_frames = bone_quat_buffer_frames_anims[name];
            auto& pos_frames = bone_pos_buffer_frames_anims[name];
            int key = 0;
            for (int frame = 0; frame < anim.frames; ++frame) {
                auto frame_begin = anim.animation.begin() + frame * anim.bones;
                auto frame_bones = boost::iterator_range<psa_animkey *>(frame_begin, frame_begin + anim.bones);
                vector<fquaternion> positions, orientations;
                calculate_skelet([&](int i) {
                    return psk_to_psa_bone[bones[psa_to_psk_bone[i]].parent_index];
                }, frame_bones, positions, orientations);

                int B = anim.bones;
                vector<float> quat(B * 4), pos(B * 4);
                for (int i = 0; i < anim.bones; ++i) {
                    int idx = psa_to_psk_bone[i];
                    quat[idx * 4 + 0] = orientations[i].R_component_2();
                    quat[idx * 4 + 1] = orientations[i].R_component_3();
                    quat[idx * 4 + 2] = orientations[i].R_component_4();
                    quat[idx * 4 + 3] = orientations[i].R_component_1();

                    pos[idx * 4 + 0] = positions[i].R_component_2();
                    pos[idx * 4 + 1] = positions[i].R_component_3();
                    pos[idx * 4 + 2] = positions[i].R_component_4();
                    pos[idx * 4 + 3] = 0.0;
                }
                quat_frames.push_back(quat);
                pos_frames.push_back(pos);
            }
        }
    }

    vector<int> psk_bones_parents()
    {
        vector<int> result;
        for (auto&& bone : bones) {
            result.push_back(bone.parent_index);
        }
        return result;
    }
};


int main()
{
    /*
    AM_DanceSexyKneeling01
    AM_DanceSexyKneeling04
    AM_DanceSexyStanding
    AM_Waiting
    HMF_WI_Dancing_Rave_01
    HMF_WI_Dancing_Rave_02
    HMF_WI_Dancing_Sexy_Pod_01
    */
    psk_psa_reader reader("c:/www/tali/QRN_ARM_TLIa_MDL.psk", "c:/www/tali/HMF_AM_Towny.psa");
    reader.prepare_buffers();
#if _DEBUG
#else
    bool binary = true;
    if (!binary) {
        cout << "var bones = " << reader.psk_bones_parents() << ";" << endl << endl;
        cout << "var vertices = " << reader.vertex_buffer << ";" << endl << endl;
        cout << "var textureCoordinates = " << reader.texcoord_buffer << ";" << endl << endl;
        cout << "var vertexIndices = " << reader.index_buffer << ";" << endl << endl;
        cout << "var boneWeight = " << reader.bone_weight_buffer << ";" << endl << endl;
        cout << "var boneData = " << reader.bone_data_buffer << ";" << endl << endl;
        cout << "var animationBonePositions = " << reader.bone_pos_buffer_frames_anims << ";" << endl << endl;
        cout << "var animationBoneOrientations = " << reader.bone_quat_buffer_frames_anims << ";" << endl << endl;
    } else {
        reader.bone_pos_buffer_frames_anims.erase("AM_DanceSexyKneeling01");
        reader.bone_pos_buffer_frames_anims.erase("AM_DanceSexyKneeling04");
        reader.bone_quat_buffer_frames_anims.erase("AM_DanceSexyKneeling01");
        reader.bone_quat_buffer_frames_anims.erase("AM_DanceSexyKneeling04");

        stringstream ss;
        ss << ultra::binaryout(reader.psk_bones_parents());
        ss << ultra::binaryout(reader.vertex_buffer);
        ss << ultra::binaryout(reader.texcoord_buffer);
        ss << ultra::binaryout(reader.index_buffer);
        ss << ultra::binaryout(reader.bone_weight_buffer);
        ss << ultra::binaryout(reader.bone_data_buffer);
        ss << ultra::binaryout(reader.bone_pos_buffer_frames_anims);
        ss << ultra::binaryout(reader.bone_quat_buffer_frames_anims);

        ofstream outf("output.bin", ios::binary);
        outf.write(ss.str().c_str(), ss.str().length());
    }
#endif
}
